use alloc::rc::Rc;
use core::cmp::Ordering;
use core::hash::{Hash, Hasher};
use core::marker::PhantomData;

#[repr(C)]
pub struct ExternRef<T> {
    p: u32,
    resource_type: PhantomData<T>,
}

impl<T> ExternRef<T> {
    pub fn is_null(&self) -> bool {
        self.p == 0
    }

    #[cfg(not(feature = "mock_data"))]
    pub(crate) fn get_internal_index(&self) -> u32 {
        self.p
    }

    #[inline]
    pub fn get_value(&self) -> Option<T>
    where
        T: ExternalReferenceType,
        T: Sized,
        T:,
    {
        T::instantiate_from_extern_value(*self)
    }

    #[inline]
    pub fn get_value_rc(&self) -> Option<Rc<T>>
    where
        T: ExternalReferenceType,
        T: Sized,
        T:,
    {
        T::instantiate_from_extern_value_rc(*self)
    }

    #[inline]
    pub fn not_null(&self) -> T
    where
        T: ExternalReferenceType,
        T: Sized,
        T:,
    {
        T::instantiate_from_extern_value(*self).unwrap()
    }

    #[inline]
    pub fn not_null_rc(&self) -> Rc<T>
    where
        T: ExternalReferenceType,
        T: Sized,
        T:,
    {
        Rc::new(T::instantiate_from_extern_value(*self).unwrap())
    }

    #[inline]
    #[cfg(not(feature = "mock_data"))]
    pub(crate) fn cast<TCast>(&self) -> ExternRef<TCast> {
        ExternRef::<TCast> {
            p: self.p,
            resource_type: Default::default(),
        }
    }
}

impl<T> Clone for ExternRef<T> {
    fn clone(&self) -> Self {
        Self {
            p: self.p,
            resource_type: Default::default(),
        }
    }
}

impl<T> Copy for ExternRef<T> {}

impl<T> Default for ExternRef<T> {
    fn default() -> Self {
        Self {
            p: 0,
            resource_type: Default::default(),
        }
    }
}

impl<T> Eq for ExternRef<T> {}

impl<T> PartialEq<Self> for ExternRef<T> {
    fn eq(&self, other: &Self) -> bool {
        u32::eq(&self.p, &other.p)
    }
}

impl<T> PartialOrd<Self> for ExternRef<T> {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        u32::partial_cmp(&self.p, &other.p)
    }
}

impl<T> Ord for ExternRef<T> {
    fn cmp(&self, other: &Self) -> Ordering {
        u32::cmp(&self.p, &other.p)
    }
}

impl<T> Hash for ExternRef<T> {
    fn hash<H: Hasher>(&self, state: &mut H) {
        self.p.hash(state)
    }
}

impl<T> From<u32> for ExternRef<T> {
    fn from(value: u32) -> Self {
        Self {
            p: value,
            resource_type: Default::default(),
        }
    }
}

pub trait ExternalReferenceType {
    fn from_extern_value(reference: ExternRef<Self>) -> Self
    where
        Self: Sized;

    fn instantiate_from_extern_value(reference: ExternRef<Self>) -> Option<Self>
    where
        Self: Sized,
    {
        if reference.is_null() {
            None
        } else {
            Some(Self::from_extern_value(reference))
        }
    }

    fn instantiate_from_extern_value_rc(reference: ExternRef<Self>) -> Option<Rc<Self>>
    where
        Self: Sized,
    {
        if reference.is_null() {
            None
        } else {
            Some(Rc::new(Self::from_extern_value(reference)))
        }
    }
}

#[macro_export]
macro_rules! impl_extern_ctor {
    ($name:ty) => {
        impl $name {
            pub(crate) fn new(ptr: ExternRef<Self>) -> Self {
                Self { ptr }
            }
            pub(crate) fn ptr(&self) -> &ExternRef<Self> {
                &self.ptr
            }
        }
    };
}
